package gin

import (
	"mime/multipart"
	"net/url"

	"github.com/spf13/cast"
)

type IRequest interface {
	// url query
	// e.g. foo.com?a=1&b=bar&c[]=bar
	QueryAll(key string) url.Values
	QueryIntWithDefault(key string, defval int) (int, bool)
	QueryInt64WithDefault(key string, defval int64) (int64, bool)
	QueryFloat32WithDefault(key string, defval float32) (float32, bool)
	QueryFloat64WithDefault(key string, defval float64) (float64, bool)
	QueryBoolWithDefault(key string, defval bool) (bool, bool)
	QueryStringWithDefault(key string, defval string) (string, bool)
	QueryStringSliceWithDefault(key string, defval []string) ([]string, bool)

	// url params
	// e.g. /book/:id
	Param(key string) string
	ParamIntWithDefault(key string, defval int) (int, bool)
	ParamInt64WithDefault(key string, defval int64) (int64, bool)
	ParamFloat32WithDefault(key string, defval float32) (float32, bool)
	ParamFloat64WithDefault(key string, defval float64) (float64, bool)
	ParamBoolWithDefault(key string, defval bool) (bool, bool)
	ParamStringWithDefault(key string, defval string) (string, bool)

	// form
	FormAllWithDefault(key string) url.Values
	FormIntWithDefault(key string, defval int) (int, bool)
	FormInt64WithDefault(key string, defval int64) (int64, bool)
	FormFloat32WithDefault(key string, defval float32) (float32, bool)
	FormFloat64WithDefault(key string, defval float64) (float64, bool)
	FormBoolWithDefault(key string, defval bool) (bool, bool)
	FormStringWithDefault(key string, defval string) (string, bool)
	FormStringSliceWithDefault(key string, defval []string) ([]string, bool)
	FormFile(key string) (*multipart.FileHeader, error)

	// JSON body
	BindJSON(obj any) error

	// XML body
	BindXML(obj any) error

	// RAW body
	GetRawData() ([]byte, error)

	// Basic informations
	Uri() string
	Method() string
	Host() string
	ClientIP() string

	// Header
	Headers() map[string][]string
	Header(key string) (string, bool)

	// Cookie
	Cookies() map[string]string
	Cookie(key string) (string, bool)
}

// {{{ url query

// QueryAll returns all queries in a request URL
func (ctx *Context) QueryAll() url.Values {
	if ctx.Request != nil {
		return map[string][]string(ctx.Request.URL.Query())
	}
	return url.Values{}
}

// QueryInt gets an int value from the query request
func (ctx *Context) QueryIntWithDefault(key string, defval int) (int, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToInt(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) QueryInt64WithDefault(key string, defval int64) (int64, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToInt64(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) QueryBoolWithDefault(key string, defval bool) (bool, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToBool(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) QueryFloat32WithDefault(key string, defval float32) (float32, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToFloat32(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) QueryFloat64WithDefault(key string, defval float64) (float64, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToFloat64(vals[0]), true
		}
	}
	return defval, false
}

// QueryString gets a string value from the query request
func (ctx *Context) QueryStringWithDefault(key string, defval string) (string, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToString(vals[0]), true
		}
	}
	return defval, false
}

// QueryArray gets an array of string values from the query request
func (ctx *Context) QueryStringSliceWithDefault(key string, defval []string) ([]string, bool) {
	params := ctx.QueryAll()
	if vals, ok := params[key]; ok {
		return cast.ToStringSlice(vals[0]), true
	}
	return defval, false
}

// }}}
// {{{ url params

func (ctx *Context) ParamIntWithDefault(key string, def int) (int, bool) {
	if val := ctx.Param(key); val != "" {
		return cast.ToInt(val), true
	}
	return def, false
}

func (ctx *Context) ParamInt64WithDefault(key string, def int64) (int64, bool) {
	if val := ctx.Param(key); val != "" {
		return cast.ToInt64(val), true
	}
	return def, false
}

func (ctx *Context) ParamFloat64WithDefault(key string, def float64) (float64, bool) {
	if val := ctx.Param(key); val != "" {
		return cast.ToFloat64(val), true
	}
	return def, false
}

func (ctx *Context) ParamFloat32WithDefault(key string, def float32) (float32, bool) {
	if val := ctx.Param(key); val != "" {
		return cast.ToFloat32(val), true
	}
	return def, false
}

func (ctx *Context) ParamBoolWithDefault(key string, def bool) (bool, bool) {
	if val := ctx.Param(key); val != "" {
		return cast.ToBool(val), true
	}
	return def, false
}

func (ctx *Context) ParamStringWithDefault(key string, def string) (string, bool) {
	if val := ctx.Param(key); val != "" {
		return cast.ToString(val), true
	}
	return def, false
}

// }}}
// {{{ Post form

// FormAll gets everything from the submitted form
func (ctx *Context) FormAll() url.Values {
	if ctx.Request != nil {
		_ = ctx.Request.ParseForm()
		return ctx.Request.PostForm
	}
	return url.Values{}
}

// FormInt gets an int value from the submitted form
func (ctx *Context) FormIntWithDefault(key string, defval int) (int, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToInt(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) FormInt64WithDefault(key string, defval int64) (int64, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToInt64(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) FormBoolWithDefault(key string, defval bool) (bool, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToBool(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) FormFloat32WithDefault(key string, defval float32) (float32, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToFloat32(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) FormFloat64WithDefault(key string, defval float64) (float64, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToFloat64(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) FormStringWithDefault(key string, defval string) (string, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		if len(vals) > 0 {
			return cast.ToString(vals[0]), true
		}
	}
	return defval, false
}

func (ctx *Context) FormStringSliceWithDefault(key string, defval []string) ([]string, bool) {
	params := ctx.FormAll()
	if vals, ok := params[key]; ok {
		return cast.ToStringSlice(vals[0]), true
	}
	return defval, false
}

// }}}
// {{{ Basic informations

func (ctx *Context) Uri() string {
	return ctx.Request.RequestURI
}

func (ctx *Context) Method() string {
	return ctx.Request.Method
}

func (ctx *Context) Host() string {
	return ctx.Request.Host
}

// }}}
// {{{ Headers

// Header
func (ctx *Context) Headers() map[string][]string {
	return ctx.Request.Header
}

// }}}
// {{{ Cookies

// Cookies gets cookie key-value pairs
func (ctx *Context) Cookies() map[string]string {
	cookies := ctx.Request.Cookies()
	ret := map[string]string{}
	for _, c := range cookies {
		ret[c.Name] = c.Value
	}
	return ret
}

// }}}
